'''
This code solves counting problem [1] using differentiable TRG programming [2] 
In a nutshell, it computes the maximum eigenvalue of tranfer matrix via variational principle
[1] Vanderstraeten et al, PRE 98, 042145 (2018)
[2] Levin and Nave, PRL 99, 120601 (2007)
'''
import torch 
torch.set_num_threads(1)
torch.manual_seed(42)

#from trg import levin_nave_trg as contraction 
from ctmrg import ctmrg as contraction 
#from vmps import vmps as contraction

def symmetrize(A):
    As = A.view(d, D, D, D, D)
    As = (As + As.permute(0, 4, 2, 3, 1))/2. 
    As = (As + As.permute(0, 1, 3, 2, 4))/2. 
    As = (As + As.permute(0, 3, 4, 1, 2))/2. 
    As = (As + As.permute(0, 2, 1, 4, 3))/2. 
    As = As.view(d, D**4)
    return As/As.norm()

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    parser.add_argument("-D", type=int, default=2, help="D")
    parser.add_argument("-Dcut", type=int, default=30, help="Dcut")
    parser.add_argument("-Niter", type=int, default=50, help="Niter")
    parser.add_argument("-Nepochs", type=int, default=10, help="Nepochs")

    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-lanczos_steps", type=int, default=0, help="lanczos steps")
    parser.add_argument("-cuda", type=int, default=-1, help="use GPU")
    args = parser.parse_args()
    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64

    if args.lanczos_steps>0: print ('lanczos steps', args.lanczos_steps)

    d = 2 # fixed

    D = args.D
    Dcut = args.Dcut
    Niter = args.Niter

    B = 0.01* torch.randn(d, D, D, D, D, dtype=dtype, device=device)
    #symmetrize initial boundary PEPS
    A = torch.nn.Parameter(symmetrize(B).view(d, D**4))
    
    #boundary MPS
    #A1 = torch.nn.Parameter(0.01*torch.randn(Dcut, D**2*d, Dcut, dtype=dtype, device=device))
    #A2 = torch.nn.Parameter(0.01*torch.randn(Dcut, D**2, Dcut, dtype=dtype, device=device))

    #dimer covering
    T = torch.zeros(d, d, d, d, d, d, dtype=dtype, device=device)
    T[0, 0, 0, 0, 0, 1] = 1.0 
    T[0, 0, 0, 0, 1, 0] = 1.0 
    T[0, 0, 0, 1, 0, 0] = 1.0 
    T[0, 0, 1, 0, 0, 0] = 1.0 
    T[0, 1, 0, 0, 0, 0] = 1.0 
    T[1, 0, 0, 0, 0, 0] = 1.0 
    T = T.view(d, d**4, d)

    optimizer = torch.optim.LBFGS([A], max_iter=20, tolerance_grad=0, tolerance_change=0, line_search_fn="strong_wolfe") 
    
    def closure():
        optimizer.zero_grad()
        As = symmetrize(A)

        T1 = torch.einsum('xa,xby,yc' , (As,T,As)).view(D,D,D,D, d,d,d,d, D,D,D,D).permute(0,4,8, 1,5,9, 2,6,10, 3,7,11).contiguous().view(D**2*d, D**2*d, D**2*d, D**2*d)
        #double layer 
        T2 = (As.t()@As).view(D, D, D, D, D, D, D, D).permute(0,4, 1,5, 2,6, 3,7).contiguous().view(D**2, D**2, D**2, D**2)
        #lnT = contraction(T1, D**2*d, Dcut, Niter, A1, lanczos_steps=args.lanczos_steps)
        #lnZ = contraction(T2, D**2, Dcut, Niter, A2, lanczos_steps=args.lanczos_steps)
        lnT, error1 = contraction(T1, D**2*d, Dcut, Niter)
        lnZ, error2 = contraction(T2, D**2, Dcut, Niter)
        loss = (-lnT + lnZ)
        #print ('    total loss', loss.item())
        #print ('    loss, error', loss.item(), error1.item(), error2.item())

        loss.backward()
        return loss
    
    for epoch in range(args.Nepochs):
        loss = optimizer.step(closure)
        print ('epoch, loss, gradnorm:', epoch, loss.item(), A.grad.norm().item())
    
    #continue with Newton-CG 
    ####################################################################3
    def fun(x, info):

        A = torch.as_tensor(x).requires_grad_()
        As = symmetrize(A)

        T1 = torch.einsum('xa,xby,yc' , (As,T,As)).view(D,D,D,D, d,d,d,d, D,D,D,D).permute(0,4,8, 1,5,9, 2,6,10, 3,7,11).contiguous().view(D**2*d, D**2*d, D**2*d, D**2*d)
        T2 = (As.t()@As).view(D, D, D, D, D, D, D, D).permute(0,4, 1,5, 2,6, 3,7).contiguous().view(D**2, D**2, D**2, D**2)
        lnT, error1 = contraction(T1, D**2*d, Dcut, Niter)
        lnZ, error2 = contraction(T2, D**2, Dcut, Niter)
        loss = (-lnT + lnZ)

        loss.backward(create_graph=True)

        info['feval'] += 1
        info['loss'] = loss
        info['A'] = A
        
        print (info['feval'], loss.item(), A.grad.norm().item())
        return loss.item(), A.grad.detach().numpy().ravel()
    
    # see this https://code.itp.ac.cn/wanglei/dpeps/commit/2e47d663bb2c8e155967c4b644edf63d9e9b22e9
    def hessp(x, p, info): 

        A = info['A']

        if ((x == A.detach().numpy().ravel()).all()):
            grad = A.grad.view(-1)
        else:
            print ('recalculate forward and grad')
            A = torch.as_tensor(A).requires_grad_()

            As = symmetrize(A)

            T1 = torch.einsum('xa,xby,yc' , (As,T,As)).view(D,D,D,D, d,d,d,d, D,D,D,D).permute(0,4,8, 1,5,9, 2,6,10, 3,7,11).contiguous().view(D**2*d, D**2*d, D**2*d, D**2*d)
            T2 = (As.t()@As).view(D, D, D, D, D, D, D, D).permute(0,4, 1,5, 2,6, 3,7).contiguous().view(D**2, D**2, D**2, D**2)
            lnT, error1 = contraction(T1, D**2*d, Dcut, Niter)
            lnZ, error2 = contraction(T2, D**2, Dcut, Niter)
            loss = (-lnT + lnZ)

            loss.backward(create_graph=True)

            grad = A.grad.view(-1)
 
        #dot it with the given vector
        loss = torch.dot(grad, torch.as_tensor(p)) 
        hvp = torch.autograd.grad(loss, A, retain_graph=True)[0].view(-1)

        return hvp.numpy().ravel()

    import scipy.optimize
    x0 = A.detach().numpy().ravel()
    x = scipy.optimize.minimize(fun, x0, args=({'feval':0},), jac=True, hessp=hessp, method='Newton-CG', options={'xtol':1E-8})
